from collections import defaultdict
from copy import copy
from itertools import combinations
from typing import Any, Collection, Hashable, Iterable, Iterator
from warnings import warn

import networkx as nx
import numpy as np
import surpyval as surv
from numpy.typing import ArrayLike

from repyability.rbd.min_path_sets import min_path_sets
from repyability.rbd.rbd_graph import RBDGraph

from .helper_classes import PerfectReliability, PerfectUnreliability
from .min_cut_sets import min_cut_sets
from .rbd_args_check import check_rbd_node_args_complete


class RBD:
    def __init__(
        self,
        nodes: dict[Any, Any],
        reliability: dict[Any, Any],
        edges: Iterable[tuple[Hashable, Hashable]],
        k: dict[Any, int] = {},
        mc_samples: int = 10_000,
    ):
        """Creates and returns a Reliability Block Diagram object.

        Parameters
        ----------
        nodes : dict[Any, Any]
            A dictionary of node names as keys and their respective component
            names as values (which map to the components in the reliability
            dict), except for the the input and output nodes which need string
            values `"input_node"` and `"output_node"` respectively
        reliability : dict[Any, Any]
            A dictionary of all non-input-output components names as keys
            with their SurPyval reliability distributions as values
        edges : Iterable[tuple[Hashable, Hashable]]
            The collection of node edges, e.g. [(1, 2), (2, 3)] would
            correspond to the edges 1-2 and 2-3
        components : dict[Any, Any]
            A dictionary of all non-input-output components names as keys
            with their SurPyval distribution as values
        mc_samples : int, optional
            TODO, by default 10_000

        Raises
        ------
        ValueError
            A node is not in the node list or edge list
        """

        # Check args are complete, will raise ValueError if not
        check_rbd_node_args_complete(nodes, reliability, edges)

        # Create RBD graph
        G = RBDGraph()
        G.add_edges_from(edges)
        self.G = G

        # Set the node k values (k-out-of-n)
        for node, k_val in k.items():
            G.nodes[node]["k"] = k_val

        # Copy the components and nodes
        reliability = copy(reliability)
        nodes = copy(nodes)

        # Look through all the nodes.
        visited_nodes = set()
        for node in nodes.keys():
            visited_nodes.add(node)

            # Set node attribute dict types if input/output
            # (if neither input/output no need to do anything, RBDGraph
            # defaults the type to "node")
            if nodes[node] == "input_node":
                self.input_node = node
                self.G.nodes[node]["type"] = "input_node"
                reliability[node] = PerfectReliability
            elif nodes[node] == "output_node":
                self.output_node = node
                self.G.nodes[node]["type"] = "output_node"
                reliability[node] = PerfectReliability

        nodes.pop(self.input_node)
        nodes.pop(self.output_node)
        self.in_or_out = [self.input_node, self.output_node]

        # Create a components to nodes dictionary for efficient sf() lookup
        self.components_to_nodes: dict[Any, set] = defaultdict(set)
        for node, component in nodes.items():
            self.components_to_nodes[component].add(node)

        # Check that all nodes in graph were in the nodes list.
        for n in G.nodes:
            if n not in visited_nodes:
                raise ValueError("Node {} not in nodes list".format(n))

        new_models = {}
        for k, v in reliability.items():
            if type(v) == list:
                sim = 0
                for model in v:
                    sim += model.random(mc_samples)

                new_models[k] = surv.KaplanMeier.fit(sim)

        # This will override the existing list with Non-Parametric
        # models
        reliability = {**reliability, **new_models}

        self.reliability = reliability
        self.nodes = nodes

        # Finally, check valid RBD structure
        if not self.is_valid_RBD_structure():
            raise ValueError(
                f"RBD not correctly structured, add edges or nodes \
                to create correct structure. See errors: \
                {self.rbd_structural_errors}"
            )

    def get_all_path_sets(self) -> Iterator[list[Hashable]]:
        """Gets all path sets from input_node to output_node

        Really just wraps networkx.all_simple_paths(). This is an expensive
        operation, so be careful using for very large RBDs.

        Returns
        -------
        Iterator[list[Hashable]]
            The iterator of paths
        """
        return nx.all_simple_paths(
            self.G, source=self.input_node, target=self.output_node
        )

    def get_min_path_sets(
        self, include_in_out_nodes=True
    ) -> set[frozenset[Hashable]]:
        """Gets the minimal path-sets of the RBD

        Parameters
        ----------
        include_in_out_nodes : bool, optional
            If false, excludes the input and output nodes
            in the return, by default True

        Returns
        -------
        set[frozenset[Hashable]]
            The set of minimal path-sets
        """
        # Run min_path_sets() but convert all the inner sets to frozensets
        # and remove the input/output nodes if requested
        ret_set = set()

        for min_path_set in min_path_sets(
            rbd_graph=self.G,
            curr_node=self.output_node,
            solns={},
        ):
            if not include_in_out_nodes:
                min_path_set.remove(self.input_node)
                min_path_set.remove(self.output_node)
            ret_set.add(frozenset(min_path_set))

        return ret_set

    def get_min_cut_sets(self) -> set[frozenset[Hashable]]:
        """
        Returns the set of frozensets of minimal cut sets of the RBD. The outer
        set contains the frozenset of nodes. frozensets were used so the inner
        set elements could be hashable.
        """
        return min_cut_sets(self.G, self.input_node, self.output_node)

    def sf(
        self,
        x: ArrayLike,
        working_nodes: Collection[Hashable] = [],
        broken_nodes: Collection[Hashable] = [],
        working_components: Collection[Hashable] = [],
        broken_components: Collection[Hashable] = [],
        using: str = "p",
    ) -> np.ndarray:
        """Returns the system reliability for time/s x.

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable
        working_nodes : Collection[Hashable], optional
            Marks these nodes as perfectly reliable, by default []
        broken_nodes : Collection[Hashable], optional
            Marks these nodes as perfectly unreliable, by default []
        working_components : Collection[Hashable], optional
            Marks these components as perfectly reliable, by default []
        broken_components : Collection[Hashable], optional
            Marks these components as perfectly unreliable, by default []
        using: str, optional
            Input either "c" or "p" for the function to use cut sets or path
            sets respectively. Defaults to path sets.

        Returns
        -------
        np.ndarray
            Reliability values for all nodes at all times x

        Raises
        ------
        ValueError
            - Working/broken node/component inconsistency (a component or node
              is supplied more than once to any of working_nodes, broken_nodes,
              working_components, broken_components)
        """
        # Check for any node/component argument inconsistency
        argument_nodes: set[object] = set()
        argument_nodes.update(working_nodes)
        argument_nodes.update(broken_nodes)
        number_of_arg_comp_nodes = 0
        for comp in working_components:
            argument_nodes.update(self.components_to_nodes[comp])
            number_of_arg_comp_nodes += len(self.components_to_nodes[comp])
        for comp in broken_components:
            argument_nodes.update(self.components_to_nodes[comp])
            number_of_arg_comp_nodes += len(self.components_to_nodes[comp])
        if len(argument_nodes) != (
            len(working_nodes) + len(broken_nodes) + number_of_arg_comp_nodes
        ):
            working_comps_nodes = [
                (comp, self.components_to_nodes[comp])
                for comp in working_components
            ]
            broken_comps_nodes = [
                (comp, self.components_to_nodes[comp])
                for comp in broken_components
            ]

            raise ValueError(
                f"Node/component inconsistency provided. i.e. you have \
                provided sf() with working/broken nodes (or respective \
                components) more than once in the arguments.\n \
                Supplied:\n\
                working_nodes: {working_nodes}\n\
                broken_nodes: {broken_nodes}\n\
                (working_components, their_nodes): {working_comps_nodes}\n\
                (broken_components, their_nodes): {broken_comps_nodes}\n"
            )

        # Turn node iterables into sets for O(1) lookup later
        working_nodes = set(working_nodes)
        broken_nodes = set(broken_nodes)

        # Per Note 3 from "UNIT 16 RELIABILITY EVALUATION OF COMPLEX SYSTEMS"
        # (https://egyankosh.ac.in/bitstream/123456789/35170/1/Unit-16.pdf),
        # you should:
        # - Apply the cut set method with only first order terms if
        #   marginal error is tolerable, and components have high reliability
        #   ?TODO: test whether indeed quicker?
        # - Or if you want to use all terms, then apply either the tie set or
        #   cut set method depending on which has less sets for the system,
        #   and thus less calculation
        #   (typically cut sets)
        #
        # At the moment, only the tie set method is implemented.

        x = np.atleast_1d(x)
        # TODO: Create cut set logic
        # Get all path sets
        paths = self.get_min_path_sets()
        num_paths = len(paths)

        # Cache all component reliabilities for efficiency
        comp_rel_cache_dict: dict[Hashable, np.ndarray] = {}
        for comp in self.reliability:
            if comp in working_components:
                comp_rel_cache_dict[comp] = PerfectReliability().sf(x)
            elif comp in broken_components:
                comp_rel_cache_dict[comp] = PerfectUnreliability().sf(x)
            else:
                comp_rel_cache_dict[comp] = self.reliability[comp].sf(x)
        # We'll just add the two 'perfect component reliabilities' to this dict
        # while we're at it, since they'll be used in lookup later
        comp_rel_cache_dict["PerfectReliability"] = PerfectReliability().sf(x)
        comp_rel_cache_dict[
            "PerfectUnreliability"
        ] = PerfectUnreliability().sf(x)

        # Perform intersection calculation, which isn't as simple as summating
        # in the case of mutual non-exclusivity
        # i is the 'level' of the intersection calc
        system_rel = np.zeros_like(x)  # Return array
        for i in range(1, num_paths + 1):
            # Get tie-set combinations for level i
            tieset_combs = combinations(paths, i)

            # Calculate the reliability of each level combination
            # Making sure to not multiply a components' reliability twice
            level_sum = np.zeros_like(x)
            for tieset_comb in tieset_combs:
                # Make a set of components out of the node path/tieset
                s = set()
                for path in tieset_comb:
                    for node in path:
                        if node in self.in_or_out:
                            continue
                        # Node working/broken takes precedence over
                        # the components reliability
                        if node in working_nodes:
                            s.add("PerfectReliability")
                        elif node in broken_nodes:
                            s.add("PerfectUnreliability")
                        else:
                            s.add(self.nodes[node])  # Add component name

                # Now calculate the tieset reliability
                tieset_rel = np.ones_like(x)
                for comp in s:
                    comp_rel = comp_rel_cache_dict[comp]
                    tieset_rel = tieset_rel * comp_rel

                # Now add the tieset reliability to the level sum
                level_sum = level_sum + tieset_rel

            # Finally add/subtract the level sum to/from the system_rel if the
            # level is even/odd
            if i % 2 == 1:
                system_rel = system_rel + level_sum
            else:
                system_rel = system_rel - level_sum

        return system_rel

    def ff(self, x: ArrayLike, *args, **kwargs) -> np.ndarray:
        """Returns the system unreliability for time/s x.

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable
        *args, **kwargs :
            Any sf() arguments

        Returns
        -------
        np.ndarray
            Uneliability values for all nodes at all times x
        """
        return 1 - self.sf(x, *args, **kwargs)

    def is_valid_RBD_structure(self) -> bool:
        """Returns False if invalid RBD structure

        Invalid RBD structure includes:
        - having cycles present
        - a non-input/output node having no in/out-nodes

        Returns
        -------
        bool
            True
        """
        has_circular_dependency = not nx.is_directed_acyclic_graph(self.G)
        node_degrees: dict = defaultdict(lambda: defaultdict(int))

        for edge in self.G.edges:
            source, target = edge
            node_degrees[source]["out"] += 1
            node_degrees[target]["in"] += 1

        input_nodes = [n for n in node_degrees.values() if n["in"] == 0]
        output_nodes = [n for n in node_degrees.values() if n["out"] == 0]
        has_node_with_no_input = len(input_nodes) != 1
        has_node_with_no_output = len(output_nodes) != 1
        if not any(
            [
                has_circular_dependency,
                has_node_with_no_input,
                has_node_with_no_output,
            ]
        ):

            self.rbd_structural_errors = None
            return True
        else:
            self.rbd_structural_errors = {
                "has_circular_dependency": has_circular_dependency,
                "has_node_with_no_input": has_node_with_no_input,
                "has_node_with_no_output": has_node_with_no_output,
            }
            return False

    # Importance measures
    # https://www.ntnu.edu/documents/624876/1277590549/chapt05.pdf/82cd565f-fa2f-43e4-a81a-095d95d39272
    def birnbaum_importance(self, x: ArrayLike) -> dict[Any, float]:
        """Returns the Birnbaum measure of importance for all nodes.

        Note: Birnbaum's measure of importance assumes all nodes are
        independent. If the RBD called on has two or more nodes associated
        with the same component then a UserWarning is raised.

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable

        Returns
        -------
        dict[Any, float]
            Dictionary with node names as keys and Birnbaum importances as
            values
        """
        for component, node_set in self.components_to_nodes.items():
            if len(node_set) > 1:
                warn(
                    f"Birnbaum's measure of importance assumes nodes are \
                     dependent, but nodes {node_set} all depend on the same \
                     component '{component}."
                )

        node_importance = {}
        for node in self.nodes.keys():
            working = self.sf(x, working_nodes=[node])
            failing = self.sf(x, broken_nodes=[node])
            node_importance[node] = working - failing
        return node_importance

    # TODO: update all importance measures to allow for component as well
    def improvement_potential(self, x: ArrayLike) -> dict[Any, float]:
        """Returns the improvement potential of all nodes.

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable

        Returns
        -------
        dict[Any, float]
            Dictionary with node names as keys and improvement potentials as
            values
        """
        node_importance = {}
        for node in self.nodes.keys():
            working = self.sf(x, working_nodes=[node])
            as_is = self.sf(x)
            node_importance[node] = working - as_is
        return node_importance

    def risk_achievement_worth(self, x: ArrayLike) -> dict[Any, float]:
        """Returns the RAW importance per Modarres & Kaminskiy. That is RAW_i =
        (unreliability of system given i failed) /
        (nominal system unreliability).

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable

        Returns
        -------
        dict[Any, float]
            Dictionary with node names as keys and RAW importances as values
        """
        node_importance = {}
        system_ff = self.ff(x)
        for node in self.nodes.keys():
            failing = self.ff(x, broken_nodes=[node])
            node_importance[node] = failing / system_ff
        return node_importance

    def risk_reduction_worth(self, x: ArrayLike) -> dict[Any, float]:
        """Returns the RRW importance per Modarres & Kaminskiy. That is RRW_i =
        (nominal unreliability of system) /
        (unreliability of system given i is working).

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable

        Returns
        -------
        dict[Any, float]
            Dictionary with node names as keys and RRW importances as values
        """
        node_importance = {}
        system_ff = self.ff(x)
        for node in self.nodes.keys():
            working = self.ff(x, working_nodes=[node])
            node_importance[node] = system_ff / working
        return node_importance

    def criticality_importance(self, x: ArrayLike) -> dict[Any, float]:
        """Returns the criticality importance of all nodes at time/s x.

        Parameters
        ----------
        x : int | float | Iterable[int  |  float]
            Time/s as a number or iterable

        Returns
        -------
        dict[Any, float]
            Dictionary with node names as keys and criticality importances as
            values
        """
        bi = self.birnbaum_importance(x)
        node_importance = {}
        system_sf = self.sf(x)
        for node in self.nodes.keys():
            node_sf = self.reliability[self.nodes[node]].sf(x)
            node_importance[node] = bi[node] * node_sf / system_sf
        return node_importance

    def fussel_vesely(
        self, x: ArrayLike, fv_type: str = "c"
    ) -> dict[Any, np.ndarray]:
        """Calculate Fussel-Vesely importance of all components at time/s x.

        Briefly, the Fussel-Vesely importance measure for node i =
        (sum of probabilities of cut-sets including node i occuring/failing) /
        (the probability of the system failing).

        Typically this measure is implemented using cut-sets as mentioned
        above, although the measure can be implemented using path-sets. Both
        are implemented here.

        fv_type dictates the method:
            "c" - cut-set
            "p" - path-set

        Parameters
        ----------
        x : ArrayLike
            Time/s as a number or iterable
        fv_type : str, optional
            Dictates the method of calculation, 'c' = cut-set and
            'p' = path-set, by default "c"

        Returns
        -------
        dict[Any, np.ndarray]
            Dictionary with node names as keys and fussel-vessely importances
            as values

        Raises
        ------
        ValueError
            TODO
        NotImplementedError
            TODO
        """
        # Get node-sets based on what method was requested
        if fv_type == "c":
            node_sets = self.get_min_cut_sets()
        elif fv_type == "p":
            node_sets = {
                frozenset(path_set)
                for path_set in self.get_min_path_sets(
                    include_in_out_nodes=False
                )
            }
        else:
            raise ValueError(
                f"fv_type must be either 'c' (cut-set) or 'p' (path-set), \
                fv_type={fv_type} was given."
            )

        # Ensure time is a numpy array
        x = np.atleast_1d(x)

        # Get system unreliability, this will be the denominator for all node
        # importance calcs
        system_unreliability = self.ff(x)

        # The return dict
        node_importance: dict[Any, np.ndarray] = {}

        # Cache the component reliabilities for efficiency
        rel_dict = {}
        for component in self.reliability.keys():
            # TODO: make log
            # Calculating reliability in the log-domain though so the
            # components' reliability can be added avoid possible underflow
            rel_dict[component] = self.reliability[component].ff(x)

        # For each node,
        for node in self.nodes.keys():
            # Sum up the probabilities of the node_sets containing the node
            # from failing
            node_fv_numerator = 0
            for node_set in node_sets:
                if node not in node_set:
                    continue
                node_set_fail_prob = 1
                # Take only the independent components in that node-set, i.e.
                # don't multiply the same component twice in a node-set
                components_in_node_set = {
                    self.nodes[fail_node] for fail_node in node_set
                }
                for component in components_in_node_set:
                    node_set_fail_prob *= rel_dict[component]
                node_fv_numerator += node_set_fail_prob

            node_importance[node] = node_fv_numerator / system_unreliability

        return node_importance

    def get_component_names(self) -> list[Hashable]:
        """Simply returns the list component names of the RBD."""
        return list(self.reliability.keys())



"""
Tests Non-Repairable Optimal Replacement Time algorithms.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
import numpy as np
import pytest
from surpyval import KaplanMeier, Weibull

from repyability.non_repairable import NonRepairable


def test_optimal_replacement():
    surv_model = Weibull.from_params((1000, 2.5))
    nr_model = NonRepairable(surv_model)

    nr_model.set_costs_planned_and_unplanned(1, 5)

    assert pytest.approx(nr_model.find_optimal_replacement(), abs=1e-1) == 493


def test_weibull_no_optimal_replacement():
    surv_model = Weibull.from_params((1000, 0.5))
    nr_model = NonRepairable(surv_model)

    nr_model.set_costs_planned_and_unplanned(1, 5)

    assert nr_model.find_optimal_replacement() == np.inf


def test_incorrect_args():
    x = Weibull.random(100, 1000, 1.5)
    np_model = KaplanMeier.fit(x)

    with pytest.raises(NotImplementedError):
        NonRepairable(np_model)

    with pytest.raises(ValueError):
        NonRepairable(1)



"""
Tests RBD's sf() method (and the briefly the ff() method).

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
import pytest

from repyability.rbd.rbd import RBD
from repyability.tests.fixed_probability import FixedProbabilityFitter


# Test sf() w/ simple series RBD
def test_rbd_sf_series(rbd_series: RBD):
    t = 5
    assert (
        pytest.approx(
            rbd_series.reliability[2].sf(t)
            * rbd_series.reliability[3].sf(t)
            * rbd_series.reliability[4].sf(t)
        )
        == rbd_series.sf(t)[0]
    )


# Test sf() w/ simple series RBD
def test_rbd_sf_parallel(rbd_parallel: RBD):
    t = 2
    assert (
        pytest.approx(
            (1 - rbd_parallel.reliability[2].sf(t))
            * (1 - rbd_parallel.reliability[3].sf(t))
            * (1 - rbd_parallel.reliability[4].sf(t))
        )
        == 1 - rbd_parallel.sf(t)[0]
    )


# Test sf() w/ a simple RBD with both parallel and series components
def test_rbd_sf_composite(rbd1: RBD):
    """Tests with an RBD with both parallel and series components."""
    t = 2
    assert (
        pytest.approx(
            (
                1
                - (1 - rbd1.reliability["pump1"].sf(t))
                * (1 - rbd1.reliability["pump2"].sf(t))
            )
            * rbd1.reliability["valve"].sf(t)
        )
        == rbd1.sf(t)[0]
    )


# Test sf() w/ an RBD that cannot be reduced to parallel or series
def test_rbd_sf_complex(rbd3: RBD):
    assert pytest.approx(0.994780625) == rbd3.sf(1000)[0]


# Test sf() w/ repeated component
# i.e. Two nodes correspond to one node
def test_rbd_sf_repeated_component(rbd_repeated_component_parallel: RBD):
    t = 2
    assert (
        pytest.approx(
            (1 - rbd_repeated_component_parallel.reliability[2].sf(t))
            * (1 - rbd_repeated_component_parallel.reliability[3].sf(t))
            * (1 - rbd_repeated_component_parallel.reliability[4].sf(t))
            # Not 'component 5' as nodes 2 and 5 are both component 2
        )
        == 1 - rbd_repeated_component_parallel.sf(t)[0]
    )


# Test sf() w/ broken node
def test_rbd_sf_broken_node(rbd_parallel: RBD):
    t = 2
    assert (
        pytest.approx(
            (1 - rbd_parallel.reliability[2].sf(t))
            * (1 - rbd_parallel.reliability[4].sf(t))
        )
        == 1 - rbd_parallel.sf(t, broken_nodes=[3])[0]
    )


# Test sf() w/ broken component
def test_rbd_sf_broken_component(rbd_repeated_component_parallel: RBD):
    t = 2
    assert (
        pytest.approx(
            (1 - rbd_repeated_component_parallel.reliability[2].sf(t))
            * (1 - rbd_repeated_component_parallel.reliability[4].sf(t))
        )
        == 1 - rbd_repeated_component_parallel.sf(t, broken_components=[3])[0]
    )
    assert (
        pytest.approx(
            (1 - rbd_repeated_component_parallel.reliability[3].sf(t))
            * (1 - rbd_repeated_component_parallel.reliability[4].sf(t))
        )
        == 1 - rbd_repeated_component_parallel.sf(t, broken_components=[2])[0]
    )


# Test sf() w/ working node
def test_rbd_sf_working_node(rbd_series: RBD):
    t = 2
    assert (
        pytest.approx(
            rbd_series.reliability[2].sf(t) * rbd_series.reliability[4].sf(t)
        )
        == rbd_series.sf(t, working_nodes=[3])[0]
    )


# Test sf() w/ working component
def test_rbd_sf_working_component(rbd_repeated_component_parallel: RBD):
    t = 2
    assert (
        pytest.approx(1)
        == rbd_repeated_component_parallel.sf(t, working_components=[2])[0]
    )
    assert (
        pytest.approx(1)
        == rbd_repeated_component_parallel.sf(t, working_components=[3])[0]
    )


# Test sf() w/ working node with repeated component
def test_rbd_sf_working_node_repeated_component(
    rbd_repeated_component_composite: RBD,
):
    rbd = rbd_repeated_component_composite
    t = 2
    assert (
        pytest.approx(1 - rbd.reliability[2].ff(t) * rbd.reliability[3].ff(t))
        == rbd.sf(t, working_nodes=[3])[0]
    )


# Test ff(), can just test parallel as ff() just calls 1 - sf()
def test_rbd_ff(rbd_parallel: RBD):
    t = 5
    assert (
        pytest.approx(
            (1 - rbd_parallel.reliability[2].sf(t))
            * (1 - rbd_parallel.reliability[3].sf(t))
            * (1 - rbd_parallel.reliability[4].sf(t))
        )
        == rbd_parallel.ff(t)[0]
    )


# Test ValueError


def test_rbd_sf_arg_inconsistency(rbd1: RBD):
    with pytest.raises(ValueError):
        rbd1.sf(2, working_nodes=["pump1"], broken_components=["pump1"])


# Test sf() w/ an RBD as a node
def test_rbd_sf_RBD_as_node(rbd1: RBD):
    """rbd_parallel but node 3 is rbd1."""
    nodes = {1: "input_node", 2: 2, 3: "RBD", 4: 4, 5: "output_node"}
    edges = [(1, 2), (1, 3), (1, 4), (2, 5), (3, 5), (4, 5)]
    components = {
        2: FixedProbabilityFitter.from_params(0.8),
        "RBD": rbd1,
        4: FixedProbabilityFitter.from_params(0.85),
    }
    rbd = RBD(nodes, components, edges)
    t = 2
    assert pytest.approx(rbd1.sf(t)) == rbd.reliability["RBD"].sf(t)
    assert (
        pytest.approx(
            rbd.reliability[2].ff(t)
            * rbd.reliability["RBD"].ff(t)
            * rbd.reliability[4].ff(t)
        )
        == rbd.ff(t)[0]
    )



"""
Tests RBD's get_min_cut_sets() method.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""

from repyability.rbd.rbd import RBD


def test_rbd_get_min_cut_sets_rbd_series(rbd_series: RBD):
    assert {
        frozenset([2]),
        frozenset([3]),
        frozenset([4]),
    } == rbd_series.get_min_cut_sets()


def test_rbd_get_min_cut_sets_rbd_parallel(rbd_parallel: RBD):
    assert {frozenset([2, 3, 4])} == rbd_parallel.get_min_cut_sets()


def test_rbd_get_min_cut_sets_rbd1(rbd1: RBD):
    assert {
        frozenset(["pump1", "pump2"]),
        frozenset(["valve"]),
    } == rbd1.get_min_cut_sets()


def test_rbd_get_min_cut_sets_rbd2(rbd2: RBD):
    assert {
        frozenset([2]),
        frozenset([3, 4]),
        frozenset([4, 5]),
        frozenset([4, 6]),
        frozenset([7]),
    } == rbd2.get_min_cut_sets()


def test_rbd_get_min_cut_sets_rbd3(rbd3: RBD):
    assert {
        frozenset([1, 3]),
        frozenset([2, 4]),
        frozenset([1, 4, 5]),
        frozenset([2, 3, 5]),
    } == rbd3.get_min_cut_sets()



import numpy as np
from scipy.integrate import quad
from scipy.optimize import minimize
from surpyval.nonparametric import NonParametric
from surpyval.parametric import Parametric


class NonRepairable:
    """
    Class to store the non-repairable information
    """

    def __init__(self, distribution):
        self.dist = distribution
        if type(distribution) == Parametric:
            self.model_parameterization = "parametric"
        elif type(distribution) == NonParametric:
            raise NotImplementedError("Non-Parametric Models not implemented")
        else:
            raise ValueError("Unknown TTF Model")

        self.cost_rate = np.vectorize(self._cost_rate)

    def set_costs_planned_and_unplanned(self, cp, cu):
        assert cp < cu
        self.cp = cp
        self.cu = cu

    def avg_replacement_time(self, t):
        out = quad(self.dist.sf, 0, t)
        return out[0]

    def _cost_rate(self, t):
        planned_costs = self.dist.sf(t) * self.cp
        unplanned_costs = (1 - self.dist.sf(t)) * self.cu
        avg_repl_time = self.avg_replacement_time(t)
        return (planned_costs + unplanned_costs) / avg_repl_time

    def _log_cost_rate(self, t):
        return np.log(self._cost_rate(t))

    def find_optimal_replacement(self, interp=None):
        if self.dist.dist.name == "Weibull":
            if self.dist.params[1] <= 1:
                return np.inf

        init = self.dist.mean()
        bounds = ((1e-8, None),)
        res = minimize(self._log_cost_rate, init, bounds=bounds, tol=1e-10)
        self.optimisation_results = res
        return res.x[0]



"""
Tests Non-Repairable Optimal Replacement Time algorithms.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
# import numpy as np
# import pytest
# from surpyval.renewal import Crow

# from repyability.repairable import Repairable

# https://reliawiki.org/index.php/Repairable_Systems_Analysis
# model = Crow.from_params([1./2.1211e-5, 1.4738])



"""
Contains function to check an RBD args are complete.
See function docstring for more info.
"""


from typing import Hashable, Iterable


def check_rbd_node_args_complete(
    nodes: dict,
    reliability: dict,
    edges: Iterable[tuple[Hashable, Hashable]],
    repairability: dict | None = None,
) -> None:
    """Checks if the RBD args are complete

    It raises a ValueError if any of the following are discovered:
    - a node is not found in the edges list
    - a node's component is not found in the reliability dict
    - a node's component is not found in the repairability dict (only checked
      if the repairability dict is provided)
    """
    # Check all nodes in edges list
    # Make a set out of the edges list for O(1) lookup
    edges_nodes: set[Hashable] = {node for edge in edges for node in edge}

    for node_name, component_name in nodes.items():
        # Check node is in edges
        if node_name not in edges_nodes:
            raise ValueError(f"Node {node_name} not in edges list.")

        # That's all that's needed for input and output nodes
        if component_name == "input_node" or component_name == "output_node":
            continue

        # Otherwise we need to check that a reliability distribution (or RBD)
        # is provided for that node, in the reliability dict
        if component_name not in reliability:
            raise ValueError(
                f"No reliability value provided for component\
                {component_name} (this component corresponds to node\
                {node_name})."
            )

        # And if repairability is provided, that a repairability distribution
        # (or RBD) is provided for that component
        if repairability is not None and component_name not in repairability:
            raise ValueError(
                f"No repairability value provided for component\
                {component_name} (this component corresponds to node\
                {node_name})."
            )



from collections import defaultdict

import numpy as np
import pytest
import surpyval as surv

from repyability.rbd.repairable_rbd import RepairableRBD


def test_repairable_rbd_missing_repairability_component():
    """
    Checks that a ValueError is raised when a component isn't provided
    with a repairability distribution.
    """
    nodes = {
        "input_node": "input_node",
        "no_repairability": "no_repairability",
        "output_node": "output_node",
    }
    reliability = {
        "no_repairability": surv.LogNormal.from_params([0.1, 0.2]),
    }
    edges = [
        ("input_node", "no_repairability"),
        ("no_repairability", "output_node"),
    ]
    repairability = {}
    with pytest.raises(ValueError):
        RepairableRBD(nodes, reliability, repairability, edges)


# One component, one simulation, tests simple case
def test_repairable_rbd_availability_one_component_1N():
    """
    The goal here is to, for only one simulation, test if the availability
    is being calculated correctly. We're achieving this by seeding the numpy
    random number generator. NOTE: This of course means that this test is
    coupled to surpyval's .random() implementation!
    """
    # Get a random seed
    seed = np.random.randint(0, 100)

    # Make a Repairable RBD which has only one non-input/output component
    # This component has an exponential distribution for both it's
    # reliability and repairability
    reliability_dist = surv.Exponential.from_params([0.25])
    repairability_dist = surv.Exponential.from_params([0.5])

    rbd = RepairableRBD(
        nodes={
            "s": "input_node",
            "c": "c",
            "t": "output_node",
        },
        edges=[
            ("s", "c"),
            ("c", "t"),
        ],
        reliability={"c": reliability_dist},
        repairability={"c": repairability_dist},
    )

    # Before anything, set the numpy seed
    np.random.seed(seed)

    # Let's run the sim for 50 units of time
    t_simulation = 50

    # Form expected availability
    # At t=0, availability=1 (it starts working)
    exp_t = [0]
    exp_availability = [1]

    # Get first failure event
    t_next_event = reliability_dist.random(1)[0]
    is_comp_working = True

    while t_next_event < t_simulation:
        if is_comp_working:
            # Failure event to be applied
            exp_t.append(t_next_event)
            exp_availability.append(0)
            is_comp_working = False

            # And get the next repair event
            # (the current time + the next event random sample)
            t_next_event += repairability_dist.random(1)[0]

        else:
            # Repair event to be applied
            exp_t.append(t_next_event)
            exp_availability.append(1)
            is_comp_working = True

            # And get the next failure event
            t_next_event += reliability_dist.random(1)[0]

    # Now check if expected == actual, remembering to reset the seed
    np.random.seed(seed)
    actual_t, actual_availability = rbd.availability(
        t_simulation=t_simulation, N=1
    )

    assert pytest.approx(exp_t) == actual_t[:-1]
    assert pytest.approx(exp_availability) == actual_availability[:-1]


# One component, ten simulations, tests the 0:N simulation loop
def test_repairable_rbd_availability_one_component_10N():
    """
    Basically a copy of test_repairable_rbd_availability_one_component_1N but
    performs 10 simulations. NOTE: Again, this is coupled with surpyval's
    .random() implementation.
    (See test_repairable_rbd_availability_one_component_1N()'s docstring.)
    """
    # Get a random seed
    seed = np.random.randint(0, 100)

    # Make a Repairable RBD which has only one non-input/output component
    # This component has an exponential distribution for both it's
    # reliability and repairability
    reliability_dist = surv.Exponential.from_params([0.25])
    repairability_dist = surv.Exponential.from_params([0.5])

    rbd = RepairableRBD(
        nodes={
            "s": "input_node",
            "c": "c",
            "t": "output_node",
        },
        edges=[
            ("s", "c"),
            ("c", "t"),
        ],
        reliability={"c": reliability_dist},
        repairability={"c": repairability_dist},
    )

    # Before anything, set the numpy seed
    np.random.seed(seed)

    # Let's run the sim for 50 units of time
    t_simulation = 50

    # Form expected availability
    # At t=0, availability=1 (it starts working)
    exp_timeline = defaultdict(lambda: 0)

    # Perform 10 simulations
    n_simulations = 10
    for i in range(n_simulations):
        # Set initial availability to 1 again
        exp_timeline[0] += 1

        # Get first failure event
        t_next_event = reliability_dist.random(1)[0]
        is_comp_working = True

        while t_next_event < t_simulation:
            if is_comp_working:
                # Failure event to be applied
                exp_timeline[t_next_event] -= 1
                is_comp_working = False

                # And get the next repair event
                # (the current time + the next event random sample)
                t_next_event += repairability_dist.random(1)[0]

            else:
                # Repair event to be applied
                exp_timeline[t_next_event] += 1
                is_comp_working = True

                # And get the next failure event
                t_next_event += reliability_dist.random(1)[0]

    # Make the time and availability lists, sorted by time
    exp_t = sorted(exp_timeline.keys())
    exp_availability = np.array([exp_timeline[t_key] for t_key in exp_t])
    exp_availability = exp_availability.cumsum() / n_simulations

    # Now check if expected == actual, remembering to reset the seed
    np.random.seed(seed)
    actual_t, actual_availability = rbd.availability(
        t_simulation=t_simulation, N=n_simulations
    )

    # Remove last element, which is the same as the second last
    assert pytest.approx(exp_t) == actual_t[:-1]
    assert pytest.approx(exp_availability) == actual_availability[:-1]


# Two parallel components, one simulation, tests the triaging of component
# events
def test_repairable_rbd_availability_two_parallel_components_1N():
    """
    We're gonna make a 2-component parallel RBD, and simulate availability
    once to test the component-event-triaging of availability(). NOTE: Again,
    this is coupled with surpyval's .random() implementation.
    (See test_repairable_rbd_availability_one_component_1N()'s docstring.)
    """

    # Make a two-component parallel RepairableRBD
    # We're just using identical exponential distributions for each for
    # simplicity
    reliability_dist = surv.Exponential.from_params([0.25])
    repairability_dist = surv.Exponential.from_params([0.5])
    rbd = RepairableRBD(
        nodes={
            "s": "input_node",
            "c1": "c1",
            "c2": "c2",
            "t": "output_node",
        },
        edges=[
            ("s", "c1"),
            ("c1", "t"),
            ("s", "c2"),
            ("c2", "t"),
        ],
        reliability={"c1": reliability_dist, "c2": reliability_dist},
        repairability={"c1": repairability_dist, "c2": repairability_dist},
    )

    # Before anything, set the numpy seed randomly
    seed = np.random.randint(0, 100)
    np.random.seed(seed)

    # Get the first failure event for each component
    next_component_events = {
        "c1": {
            "component": "c1",
            "type": "failure",
            "time": reliability_dist.random(1)[0],
        },
        "c2": {
            "component": "c2",
            "type": "failure",
            "time": reliability_dist.random(1)[0],
        },
    }

    # Keep track of component states, they both start working
    component_states = {"c1": True, "c2": True}

    # Helper function to determine from component_states if system is working
    def is_system_working():
        return component_states["c1"] or component_states["c2"]

    # Helper function to get the next event from next_component_events
    def get_next_event():
        if (
            next_component_events["c1"]["time"]
            < next_component_events["c2"]["time"]
        ):
            return next_component_events["c1"]
        else:
            return next_component_events["c2"]

    # Let's run the sim for 50 units of time
    t_simulation = 50

    # Keep track of the system's state, it starts off working
    prev_system_state = True

    # Form expected availability
    # At t=0, availability=1 (it starts working)
    exp_t = [0]
    exp_availability = [1]

    # Get the first failure event
    next_event = get_next_event()

    # While the next event's time is in the simulation
    while next_event["time"] < t_simulation:
        # This event's component
        component = next_event["component"]

        # If the next event is a failure, set the component's state to not
        # working, and if the system was previously working, but is now
        # not working, record this in exp_availability, and add a new
        # repair event to next_component_states
        if next_event["type"] == "failure":
            component_states[component] = False
            if prev_system_state and not is_system_working():
                exp_t.append(next_event["time"])
                exp_availability.append(0)
                prev_system_state = False

            next_component_events[component] = {
                "component": component,
                "type": "repair",
                "time": next_event["time"] + repairability_dist.random(1)[0],
            }

        # Else, the next event is a repair, set the component's state to
        # working, and if the system was previously not working, but is now
        # working, record this in exp_availability, and add a new
        # failure event to next_component_states
        else:
            component_states[component] = True
            if not prev_system_state and is_system_working():
                exp_t.append(next_event["time"])
                exp_availability.append(1)
                prev_system_state = True

            next_component_events[component] = {
                "component": component,
                "type": "failure",
                "time": next_event["time"] + reliability_dist.random(1)[0],
            }

        # Get next event
        next_event = get_next_event()

    # Now check if expected == actual, remembering to reset the seed
    np.random.seed(seed)
    actual_t, actual_availability = rbd.availability(
        t_simulation=t_simulation, N=1
    )

    assert pytest.approx(exp_t) == actual_t[:-1]
    assert pytest.approx(exp_availability) == actual_availability[:-1]


# Two series components, one simulation, tests the triaging of component events
def test_repairable_rbd_availability_two_series_components_1N():
    """
    We're gonna make a 2-component series RBD, and simulate availability
    once to test the component-event-triaging of availability(). NOTE: Again,
    this is coupled with surpyval's .random() implementation.
    (See test_repairable_rbd_availability_one_component_1N()'s docstring.)
    """

    # Make a two-component series RepairableRBD
    # We're just using identical exponential distributions for each for
    # simplicity
    reliability_dist = surv.Exponential.from_params([0.25])
    repairability_dist = surv.Exponential.from_params([0.5])
    rbd = RepairableRBD(
        nodes={
            "s": "input_node",
            "c1": "c1",
            "c2": "c2",
            "t": "output_node",
        },
        edges=[
            ("s", "c1"),
            ("c1", "c2"),
            ("c2", "t"),
        ],
        reliability={"c1": reliability_dist, "c2": reliability_dist},
        repairability={"c1": repairability_dist, "c2": repairability_dist},
    )

    # Before anything, set the numpy seed randomly
    seed = np.random.randint(0, 100)
    np.random.seed(seed)

    # Get the first failure event for each component
    next_component_events = {
        "c1": {
            "component": "c1",
            "type": "failure",
            "time": reliability_dist.random(1)[0],
        },
        "c2": {
            "component": "c2",
            "type": "failure",
            "time": reliability_dist.random(1)[0],
        },
    }

    # Keep track of component states, they both start working
    component_states = {"c1": True, "c2": True}

    # Helper function to determine from component_states if system is working
    def is_system_working():
        return component_states["c1"] and component_states["c2"]

    # Helper function to get the next event from next_component_events
    def get_next_event():
        if (
            next_component_events["c1"]["time"]
            < next_component_events["c2"]["time"]
        ):
            return next_component_events["c1"]
        else:
            return next_component_events["c2"]

    # Let's run the sim for 50 units of time
    t_simulation = 50

    # Keep track of the system's state, it starts off working
    prev_system_state = True

    # Form expected availability
    # At t=0, availability=1 (it starts working)
    exp_t = [0]
    exp_availability = [1]

    # Get the first failure event
    next_event = get_next_event()

    # While the next event's time is in the simulation
    while next_event["time"] < t_simulation:
        # This event's component
        component = next_event["component"]

        # If the next event is a failure, set the component's state to not
        # working, and if the system was previously working, but is now
        # not working, record this in exp_availability, and add a new
        # repair event to next_component_states
        if next_event["type"] == "failure":
            component_states[component] = False
            if prev_system_state and not is_system_working():
                exp_t.append(next_event["time"])
                exp_availability.append(0)
                prev_system_state = False

            next_component_events[component] = {
                "component": component,
                "type": "repair",
                "time": next_event["time"] + repairability_dist.random(1)[0],
            }

        # Else, the next event is a repair, set the component's state to
        # working, and if the system was previously not working, but is now
        # working, record this in exp_availability, and add a new
        # failure event to next_component_states
        else:
            component_states[component] = True
            if not prev_system_state and is_system_working():
                exp_t.append(next_event["time"])
                exp_availability.append(1)
                prev_system_state = True

            next_component_events[component] = {
                "component": component,
                "type": "failure",
                "time": next_event["time"] + reliability_dist.random(1)[0],
            }

        # Get next event
        next_event = get_next_event()

    # Now check if expected == actual, remembering to reset the seed
    np.random.seed(seed)
    actual_t, actual_availability = rbd.availability(
        t_simulation=t_simulation, N=1
    )

    assert pytest.approx(exp_t) == actual_t[:-1]
    assert pytest.approx(exp_availability) == actual_availability[:-1]



import numpy as np


class FixedProbability:
    def sf(self, x):
        return np.ones_like(x) * self.p

    def ff(self, x):
        return 1 - (np.ones_like(x) * self.p)


class FixedProbabilityFitter:
    @classmethod
    def from_params(cls, p):
        out = FixedProbability()
        out.p = p
        return out



"""
Contains the pytest fixtures used by many of the tests. They all define
and return RBDs.
"""

import pytest
import surpyval as surv

from repyability.rbd.rbd import RBD
from repyability.tests.fixed_probability import FixedProbabilityFitter


@pytest.fixture
def rbd_series() -> RBD:
    """A simple RBD with three intermediate nodes in series."""
    nodes = {1: "input_node", 2: 2, 3: 3, 4: 4, 5: "output_node"}
    edges = [(1, 2), (2, 3), (3, 4), (4, 5)]
    reliability = {
        2: surv.Weibull.from_params([20, 2]),
        3: surv.Weibull.from_params([100, 3]),
        4: surv.Weibull.from_params([50, 20]),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd_parallel() -> RBD:
    """A simple RBD with three intermediate nodes in parallel."""
    nodes = {1: "input_node", 2: 2, 3: 3, 4: 4, 5: "output_node"}
    edges = [(1, 2), (1, 3), (1, 4), (2, 5), (3, 5), (4, 5)]
    reliability = {
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd1() -> RBD:
    """Example 6.10 from Modarres & Kaminskiy."""
    qp = 0.03
    qv = 0.01
    nodes = {
        "source": "input_node",
        "pump1": "pump1",
        "pump2": "pump2",
        "valve": "valve",
        "sink": "output_node",
    }
    edges = [
        ("source", "pump1"),
        ("source", "pump2"),
        ("pump1", "valve"),
        ("pump2", "valve"),
        ("valve", "sink"),
    ]
    reliability = {
        "pump1": FixedProbabilityFitter.from_params(1 - qp),
        "pump2": FixedProbabilityFitter.from_params(1 - qp),
        "valve": FixedProbabilityFitter.from_params(1 - qv),
    }

    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd2() -> RBD:
    edges = [(1, 2), (2, 3), (2, 4), (4, 7), (3, 5), (5, 6), (6, 7), (7, 8)]
    nodes = {
        1: "input_node",
        8: "output_node",
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        6: 6,
        7: 7,
    }
    reliability = {
        2: surv.Weibull.from_params([20, 2]),
        3: surv.Weibull.from_params([100, 3]),
        4: surv.Weibull.from_params([50, 20]),
        5: surv.Weibull.from_params([15, 1.2]),
        6: surv.Weibull.from_params([80, 10]),
        7: [
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
        ],
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd3() -> RBD:
    """
    Fig. 16.1 from "UNIT 16 RELIABILITY EVALUATION OF COMPLEX SYSTEMS" by
    ignou (https://egyankosh.ac.in/bitstream/123456789/35170/1/Unit-16.pdf).
    """
    nodes = {0: "input_node", 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: "output_node"}
    edges = [
        (0, 1),
        (0, 3),
        (1, 2),
        (3, 4),
        (1, 5),
        (3, 5),
        (5, 2),
        (5, 4),
        (2, 6),
        (4, 6),
    ]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.95),
        2: FixedProbabilityFitter.from_params(0.95),
        3: FixedProbabilityFitter.from_params(0.95),
        4: FixedProbabilityFitter.from_params(0.95),
        5: FixedProbabilityFitter.from_params(0.95),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd_repeated_component_parallel() -> RBD:
    """Basically rbd_parallel with a repeated component (component 2)."""
    nodes = {1: "input_node", 2: 2, 3: 3, 4: 4, 5: 2, 6: "output_node"}
    edges = [(1, 2), (1, 3), (1, 4), (1, 5), (2, 6), (3, 6), (4, 6), (5, 6)]
    reliability = {
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd_repeated_component_series() -> RBD:
    """A simple RBD with three intermediate nodes in series."""
    nodes = {1: "input_node", 2: 2, 3: 3, 4: 2, 5: "output_node"}
    edges = [(1, 2), (2, 3), (3, 4), (4, 5)]
    reliability = {
        2: surv.Weibull.from_params([20, 2]),
        3: surv.Weibull.from_params([100, 3]),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd_repeated_component_composite() -> RBD:
    """
    An RBD with three intermediate nodes, two of them a repeated component.
    """
    nodes = {1: "input_node", 2: 2, 3: 2, 4: 3, 5: "output_node"}
    edges = [(1, 2), (1, 3), (2, 5), (3, 4), (4, 5)]
    reliability = {
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.5),
    }
    return RBD(nodes, reliability, edges)


@pytest.fixture
def rbd_series_koon() -> RBD:
    """
    Series RBD w/ three intermediate nodes, the middle one with k=2.
    i.e. this system doesn't work at all.
    """
    nodes = {1: "input_node", 2: 2, 3: 3, 4: 4, 5: "output_node"}
    edges = [(1, 2), (2, 3), (3, 4), (4, 5)]
    reliability = {
        2: surv.Weibull.from_params([20, 2]),
        3: surv.Weibull.from_params([100, 3]),
        4: surv.Weibull.from_params([50, 20]),
    }
    k = {3: 2}
    return RBD(nodes, reliability, edges, k)


@pytest.fixture
def rbd1_koon() -> RBD:
    """Example 6.10 from Modarres & Kaminskiy, w/ valve k=2."""
    qp = 0.03
    qv = 0.01
    nodes = {
        "source": "input_node",
        "pump1": "pump1",
        "pump2": "pump2",
        "valve": "valve",
        "sink": "output_node",
    }
    edges = [
        ("source", "pump1"),
        ("source", "pump2"),
        ("pump1", "valve"),
        ("pump2", "valve"),
        ("valve", "sink"),
    ]
    reliability = {
        "pump1": FixedProbabilityFitter.from_params(1 - qp),
        "pump2": FixedProbabilityFitter.from_params(1 - qp),
        "valve": FixedProbabilityFitter.from_params(1 - qv),
    }
    k = {"valve": 2}
    return RBD(nodes, reliability, edges, k)


@pytest.fixture
def rbd2_koon() -> RBD:
    """rbd2 but k of node 7 is =2."""
    edges = [(1, 2), (2, 3), (2, 4), (4, 7), (3, 5), (5, 6), (6, 7), (7, 8)]
    nodes = {
        1: "input_node",
        8: "output_node",
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        6: 6,
        7: 7,
    }
    reliability = {
        2: surv.Weibull.from_params([20, 2]),
        3: surv.Weibull.from_params([100, 3]),
        4: surv.Weibull.from_params([50, 20]),
        5: surv.Weibull.from_params([15, 1.2]),
        6: surv.Weibull.from_params([80, 10]),
        7: [
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
            surv.Weibull.from_params([5, 1.1]),
        ],
    }
    k = {7: 2}
    return RBD(nodes, reliability, edges, k)


def rbd3_nodes_edges_components():
    nodes = {0: "input_node", 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: "output_node"}
    edges = [
        (0, 1),
        (0, 3),
        (1, 2),
        (3, 4),
        (1, 5),
        (3, 5),
        (5, 2),
        (5, 4),
        (2, 6),
        (4, 6),
    ]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.95),
        2: FixedProbabilityFitter.from_params(0.95),
        3: FixedProbabilityFitter.from_params(0.95),
        4: FixedProbabilityFitter.from_params(0.95),
        5: FixedProbabilityFitter.from_params(0.95),
    }
    return nodes, edges, reliability


@pytest.fixture
def rbd3_koon1() -> RBD:
    """
    Fig. 16.1 from "UNIT 16 RELIABILITY EVALUATION OF COMPLEX SYSTEMS" by
    ignou (https://egyankosh.ac.in/bitstream/123456789/35170/1/Unit-16.pdf).
    w/ k(5)=2
    """
    nodes, edges, reliability = rbd3_nodes_edges_components()
    k = {5: 2}
    return RBD(nodes, reliability, edges, k)


@pytest.fixture
def rbd3_koon2() -> RBD:
    """
    Fig. 16.1 from "UNIT 16 RELIABILITY EVALUATION OF COMPLEX SYSTEMS" by
    ignou (https://egyankosh.ac.in/bitstream/123456789/35170/1/Unit-16.pdf).
    w/ k(2)=2
    """
    nodes, edges, reliability = rbd3_nodes_edges_components()
    k = {2: 2}
    return RBD(nodes, reliability, edges, k)


@pytest.fixture
def rbd3_koon3() -> RBD:
    """
    Fig. 16.1 from "UNIT 16 RELIABILITY EVALUATION OF COMPLEX SYSTEMS" by
    ignou (https://egyankosh.ac.in/bitstream/123456789/35170/1/Unit-16.pdf).
    w/ k(2)=k(5)=2
    """
    nodes, edges, reliability = rbd3_nodes_edges_components()
    k = {2: 2, 5: 2}
    return RBD(nodes, reliability, edges, k)


@pytest.fixture
def rbd_koon_parallel_args() -> dict:
    """
    A s-3-1-t RBD.
    Returns the {nodes, edges, reliability} dict, useful so k(4) can be
    changed.
    """
    nodes = {"s": "input_node", 1: 1, 2: 2, 3: 3, "v": 4, "t": "output_node"}
    edges = [
        ("s", 1),
        ("s", 2),
        ("s", 3),
        (1, "v"),
        (2, "v"),
        (3, "v"),
        ("v", "t"),
    ]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.85),
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
    }
    return {"nodes": nodes, "edges": edges, "reliability": reliability}


@pytest.fixture
def rbd_koon_composite_args() -> dict:
    """
    A s-2-1-2-1-t RBD.
    Returns the {nodes, edges, reliability} dict, useful so k("v1") and k("v2")
    can be changed.
    """
    nodes = {
        "s": "input_node",
        "a1": 1,
        "a2": 2,
        "v1": 3,
        "b1": 4,
        "b2": 5,
        "v2": 6,
        "t": "output_node",
    }
    edges = [
        ("s", "a1"),
        ("s", "a2"),
        ("a1", "v1"),
        ("a2", "v1"),
        ("v1", "b1"),
        ("v1", "b2"),
        ("b1", "v2"),
        ("b2", "v2"),
        ("v2", "t"),
    ]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.85),
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
        5: FixedProbabilityFitter.from_params(0.85),
        6: FixedProbabilityFitter.from_params(0.8),
    }
    return {"nodes": nodes, "edges": edges, "reliability": reliability}


@pytest.fixture
def rbd_koon_simple() -> RBD:
    nodes = {"s": "input_node", 1: 1, 2: 2, "t": "output_node"}
    edges = [("s", 1), (1, 2), ("s", 2), (2, "t")]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.85),
        2: FixedProbabilityFitter.from_params(0.8),
    }
    return RBD(nodes, reliability, edges, {2: 2})


@pytest.fixture
def rbd_koon_nonminimal_args() -> dict:
    """RBD that can easily trap an algorithm into including a non-minimal
    path-set."""
    nodes = {
        "s": "input_node",
        1: 1,
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        "t": "output_node",
    }
    edges = [("s", 1), (1, 2), (2, "t"), (1, 3), (3, 4), (4, 5), (5, 2)]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.85),
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
        5: FixedProbabilityFitter.from_params(0.85),
    }
    return {"nodes": nodes, "edges": edges, "reliability": reliability}


@pytest.fixture
def rbd_double_parallel_args() -> dict:
    """RBD arguments that make for a 1-2-1-2-1 RBD."""
    nodes = {
        "s": "input_node",
        1: 1,
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        "t": "output_node",
    }
    edges = [
        ("s", 1),
        ("s", 2),
        (1, 3),
        (2, 3),
        (3, 4),
        (3, 5),
        (4, "t"),
        (5, "t"),
    ]
    reliability = {
        1: FixedProbabilityFitter.from_params(0.85),
        2: FixedProbabilityFitter.from_params(0.8),
        3: FixedProbabilityFitter.from_params(0.9),
        4: FixedProbabilityFitter.from_params(0.85),
        5: FixedProbabilityFitter.from_params(0.85),
    }
    return {"nodes": nodes, "edges": edges, "reliability": reliability}



"""
Tests many of the RBD class methods.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""

from repyability.rbd.rbd import RBD


def test_rbd_koon_default_k(rbd_series: RBD):
    assert rbd_series.G.nodes[1]["k"] == 1
    assert rbd_series.G.nodes[2]["k"] == 1
    assert rbd_series.G.nodes[3]["k"] == 1
    assert rbd_series.G.nodes[4]["k"] == 1
    assert rbd_series.G.nodes[5]["k"] == 1


def test_rbd_koon_k_given(rbd1_koon: RBD):
    assert rbd1_koon.G.nodes["source"]["k"] == 1
    assert rbd1_koon.G.nodes["pump1"]["k"] == 1
    assert rbd1_koon.G.nodes["pump2"]["k"] == 1
    assert rbd1_koon.G.nodes["valve"]["k"] == 2
    assert rbd1_koon.G.nodes["sink"]["k"] == 1


def test_rbd_get_min_path_sets_rbd_series_koon(rbd_series_koon: RBD):
    assert set() == rbd_series_koon.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd1_koon(rbd1_koon: RBD):
    assert {
        frozenset(["source", "pump1", "pump2", "valve", "sink"])
    } == rbd1_koon.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd1_exclude_in_out_node_koon(rbd1_koon: RBD):
    assert {
        frozenset(["pump1", "pump2", "valve"])
    } == rbd1_koon.get_min_path_sets(include_in_out_nodes=False)


def test_rbd_get_min_path_sets_rbd2_koon(rbd2_koon: RBD):
    assert {
        frozenset(frozenset([1, 2, 3, 4, 5, 6, 7, 8])),
    } == rbd2_koon.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd3_koon1(rbd3_koon1: RBD):
    assert {
        frozenset([0, 1, 2, 6]),
        frozenset([0, 3, 4, 6]),
    } == rbd3_koon1.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd3_koon2(rbd3_koon2: RBD):
    assert {
        frozenset([0, 1, 2, 5, 6]),
        frozenset([0, 1, 5, 4, 6]),
        frozenset([0, 3, 4, 6]),
    } == rbd3_koon2.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd3_koon3(rbd3_koon3: RBD):
    assert {
        frozenset([0, 1, 2, 3, 5, 6]),
        frozenset({0, 3, 4, 6}),
    } == rbd3_koon3.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_parallel_k1(rbd_koon_parallel_args):
    rbd = RBD(**rbd_koon_parallel_args, k={"v": 1})
    assert {
        frozenset(["s", 1, "v", "t"]),
        frozenset(["s", 2, "v", "t"]),
        frozenset(["s", 3, "v", "t"]),
    } == rbd.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_parallel_k2(rbd_koon_parallel_args):
    rbd = RBD(**rbd_koon_parallel_args, k={"v": 2})
    assert {
        frozenset(["s", 1, 2, "v", "t"]),
        frozenset(["s", 1, 3, "v", "t"]),
        frozenset(["s", 2, 3, "v", "t"]),
    } == rbd.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_parallel_k3(rbd_koon_parallel_args):
    rbd = RBD(**rbd_koon_parallel_args, k={"v": 3})
    assert {
        frozenset(["s", 1, 2, 3, "v", "t"]),
    } == rbd.get_min_path_sets()


# There are 4 combinations of (k("v1"), k("v2")) for the rbd_koon_composite
# RBD, let's test all of them...
def test_rbd_get_min_path_sets_rbd_koon_composite_k11(rbd_koon_composite_args):
    rbd = RBD(**rbd_koon_composite_args, k={"v1": 1, "v2": 1})
    assert {
        frozenset(["s", "a1", "v1", "b1", "v2", "t"]),
        frozenset(["s", "a1", "v1", "b2", "v2", "t"]),
        frozenset(["s", "a2", "v1", "b1", "v2", "t"]),
        frozenset(["s", "a2", "v1", "b2", "v2", "t"]),
    } == rbd.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_composite_k12(rbd_koon_composite_args):
    rbd = RBD(**rbd_koon_composite_args, k={"v1": 1, "v2": 2})
    assert {
        frozenset(["s", "a1", "v1", "b1", "b2", "v2", "t"]),
        frozenset(["s", "a2", "v1", "b1", "b2", "v2", "t"]),
    } == rbd.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_composite_k21(rbd_koon_composite_args):
    rbd = RBD(**rbd_koon_composite_args, k={"v1": 2, "v2": 1})
    assert {
        frozenset(["s", "a1", "a2", "v1", "b1", "v2", "t"]),
        frozenset(["s", "a1", "a2", "v1", "b2", "v2", "t"]),
    } == rbd.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_composite_k22(rbd_koon_composite_args):
    rbd = RBD(**rbd_koon_composite_args, k={"v1": 2, "v2": 2})
    assert {
        frozenset(["s", "a1", "a2", "v1", "b1", "b2", "v2", "t"]),
    } == rbd.get_min_path_sets()


# And test impossible cases!


def test_rbd_get_min_path_sets_rbd_koon_composite_nonfunctioning(
    rbd_koon_composite_args,
):
    assert (
        set()
        == RBD(
            **rbd_koon_composite_args, k={"v1": 2, "v2": 3}
        ).get_min_path_sets()
    )
    assert (
        set()
        == RBD(
            **rbd_koon_composite_args, k={"v1": 3, "v2": 2}
        ).get_min_path_sets()
    )
    assert (
        set()
        == RBD(
            **rbd_koon_composite_args, k={"v1": 3, "v2": 3}
        ).get_min_path_sets()
    )


def test_rbd_get_min_path_sets_rbd_koon_simple(rbd_koon_simple: RBD):
    assert {frozenset(["s", 1, 2, "t"])} == rbd_koon_simple.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_koon_nonminimal(
    rbd_koon_nonminimal_args: dict,
):
    rbd_k1 = RBD(**rbd_koon_nonminimal_args, k={2: 1})
    assert {frozenset(["s", 1, 2, "t"])} == rbd_k1.get_min_path_sets()

    rbd_k2 = RBD(**rbd_koon_nonminimal_args, k={2: 2})
    assert {frozenset(["s", 1, 2, 3, 4, 5, "t"])} == rbd_k2.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_double_parallel_koon(
    rbd_double_parallel_args: dict,
):
    rbd = RBD(**rbd_double_parallel_args, k={3: 2})
    assert {
        frozenset(["s", 1, 2, 3, 4, "t"]),
        frozenset(["s", 1, 2, 3, 5, "t"]),
    } == rbd.get_min_path_sets()



import copy
import itertools
from typing import Hashable

from repyability.rbd.rbd_graph import RBDGraph


def min_path_sets(
    rbd_graph: RBDGraph,
    curr_node: Hashable,
    solns: dict[Hashable, list[set]],
) -> list[set[Hashable]]:
    """Returns the minimal path sets for an RBDGraph up until curr_node

    This is used as a recursive function in a memoised approach to solving
    the k-out-of-n minimal path-set problem.

    Parameters
    ----------
    rbd_graph : RBDGraph
        The complete RBDGraph
    curr_node : Hashable
        The node to 'go up to', the first call should be the output_node
    solns : dict[Hashable, list[set]]
        The 'memo table' to prevent solving the same subproblem

    Returns
    -------
    list[set[Hashable]]
        The list of minimal path-sets, empty if no path-sets can be made
        (including consideration of the k-out-of-n behaviour)
    """

    # Base case: if curr_node is the input_node then return the input_node
    if rbd_graph.nodes[curr_node]["type"] == "input_node":
        ret_list = [{curr_node}]
        # Store in the solns dict, and return
        solns[curr_node] = copy.deepcopy(ret_list)
        return ret_list

    # Else, solve the subproblems
    # all_subproblem_solns is like [[subproblem1_sets],[subproblem2_sets],...]
    all_subproblem_solns: list[list[set[Hashable]]] = []
    for in_node in rbd_graph.predecessors(curr_node):
        # Get the subproblem soln (a list of minimal path-sets)
        if in_node in solns:
            # If it's already been solved, just used the memoised result
            # Needs to be deepcopied so it doesn't affect solns[in_node]
            subproblem_soln = copy.deepcopy(solns[in_node])
        else:
            # Otherwise go ahead and solve the subproblem
            subproblem_soln = min_path_sets(
                rbd_graph,
                in_node,
                solns,
            )

        # Add curr_node to each set
        for s in subproblem_soln:
            s.add(curr_node)

        # Collect the subproblem solns, if there are any
        # (subproblem_soln is non-empty)
        # We don't want to add an empty list to all_subproblem_solns
        # as it makes it simpler for the later steps
        if subproblem_soln:
            all_subproblem_solns.append(subproblem_soln)

    # Get the curr_node's k value
    k = rbd_graph.nodes[curr_node]["k"]

    # If there's no way of satisfying the k requirements of curr_node,
    # just skip all the funny business and return the empty list
    if len(all_subproblem_solns) < k:
        solns[curr_node] = []
        return []

    # Merge the subproblem solns to satisfy the curr_node's k requirement
    # e.g. k = 1, just add all the sets to the ret_list
    # e.g. k = 2, add all possible pairs across subproblems
    #   i.e. if there's 2 subproblems, each with 2 minimal path-sets
    #        (aa, ab, ba, bb) then merge the sets (aa+ba, aa+bb, ab+ba, ab+bb)

    # Get all the necessary path-set combinations
    merged_soln = merge_subproblem_solns(all_subproblem_solns, k)

    # Remove all non-minimal path-sets
    merged_and_minimal_soln = minimalise_path_sets(merged_soln)

    # Finally, memoise the solution to this curr_node problem, and return
    if rbd_graph.out_degree(curr_node) > 1:
        solns[curr_node] = copy.deepcopy(merged_and_minimal_soln)

    return merged_and_minimal_soln


def get_node_in_combinations(n_in: int, k: int) -> list[tuple[int, ...]]:
    """Returns all the indexed-0 node combinations for a k k-out-of-n node with
    n_in predecessors

    e.g. If the k-out-of-n node has k=2, and has 4 predecessors, this will
    return [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)].
    """
    return list(itertools.combinations(range(n_in), k))


def merge_subproblem_solns(
    all_subproblem_solns: list[list[set[Hashable]]], curr_node_k: int
) -> list[set[Hashable]]:
    """Returns k-wise path-set combinations for the current node"""
    # First get the node-in combinations
    n_in = len(all_subproblem_solns)
    combs = get_node_in_combinations(n_in, curr_node_k)

    # Then form the path-set combinations (or set-additions rather)

    merged_soln: list[set] = []
    for comb in combs:
        # product_candidates are all the subproblem solns we're considering for
        # this combination
        product_candidates: list[list[set]] = [
            all_subproblem_solns[i] for i in comb
        ]

        # itertools.product() performs all the magic for us, basically
        # getting us the '(aa, ab, ba, bb)' we need
        for product_tup in itertools.product(*product_candidates):
            # Now merge the sets, i.e. ({1, 2}, {3, 4}) -> {1, 2, 3, 4}
            s = set(product_tup[0])
            for i in range(1, len(product_tup)):
                s |= product_tup[i]

            # And add it to the merged_soln list
            merged_soln.append(s)

    return merged_soln


def minimalise_path_sets(path_sets: list[set]) -> list[set]:
    """Returns only the minimal path-sets

    i.e. Discards all path-sets that are a superset of any of the other
    path-sets.

    Parameters
    ----------
    path_sets : list[set]
        All path-sets to minimalise
    """
    minimal_path_sets = []
    for path_set in path_sets:
        is_path_set_minimal = True  # Assume path-set is minimal
        for other_path_set in path_sets:
            if path_set == other_path_set:
                # Don't consider the same path-set
                continue
            if path_set.issuperset(other_path_set):
                # Discard path_set as it's non-minimal (other_path_set is a
                # subset of this path_set)
                is_path_set_minimal = False
                break
        if is_path_set_minimal:
            minimal_path_sets.append(path_set)
    return minimal_path_sets



"""
Tests RBD's importance methods.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
import pytest

from repyability.rbd.rbd import RBD


# Test birnbaum_importance() w/ composite RBD
def test_rbd_birnbaum_importance(rbd1: RBD):
    t = 2
    birnbaum_importance_dict = rbd1.birnbaum_importance(t)
    assert len(birnbaum_importance_dict) == 3
    assert (
        pytest.approx(
            rbd1.reliability["valve"].sf(t)
            - rbd1.reliability["pump2"].sf(t) * rbd1.reliability["valve"].sf(t)
        )
        == birnbaum_importance_dict["pump1"]
    )
    assert (
        pytest.approx(
            rbd1.reliability["valve"].sf(t)
            - rbd1.reliability["pump1"].sf(t) * rbd1.reliability["valve"].sf(t)
        )
        == birnbaum_importance_dict["pump2"]
    )
    assert (
        pytest.approx(
            1
            - rbd1.reliability["pump1"].ff(t) * rbd1.reliability["pump2"].ff(t)
        )
        == birnbaum_importance_dict["valve"]
    )


# Test birnbaum_importance() raises warning when dependence is found
def test_rbd_birnbaum_importance_dependence_warning(
    rbd_repeated_component_series: RBD,
):
    with pytest.warns(UserWarning):
        rbd_repeated_component_series.birnbaum_importance(2)


# Test improvement_potential()
def test_rbd_improvement_potential(rbd1: RBD):
    t = 2
    improvement_potential = rbd1.improvement_potential(t)
    assert len(improvement_potential) == 3
    assert (
        pytest.approx(rbd1.reliability["valve"].sf(t) - rbd1.sf(t))
        == improvement_potential["pump1"]
    )
    assert (
        pytest.approx(rbd1.reliability["valve"].sf(t) - rbd1.sf(t))
        == improvement_potential["pump2"]
    )
    assert (
        pytest.approx(
            (
                1
                - rbd1.reliability["pump1"].ff(t)
                * rbd1.reliability["pump2"].ff(t)
            )
            - rbd1.sf(t)
        )
        == improvement_potential["valve"]
    )


# Test risk_achievement_worth()
def test_rbd_risk_achievement_worth(rbd1: RBD):
    t = 2
    raw = rbd1.risk_achievement_worth(t)
    assert len(raw) == 3
    assert (
        pytest.approx(
            (
                1
                - rbd1.reliability["pump2"].sf(t)
                * rbd1.reliability["valve"].sf(t)
            )
            / rbd1.ff(t)
        )
        == raw["pump1"]
    )
    assert (
        pytest.approx(
            (
                1
                - rbd1.reliability["pump1"].sf(t)
                * rbd1.reliability["valve"].sf(t)
            )
            / rbd1.ff(t)
        )
        == raw["pump2"]
    )
    assert pytest.approx(1 / rbd1.ff(t)) == raw["valve"]


# Test risk_reduction_worth()
def test_rbd_risk_reduction_worth(rbd1: RBD):
    t = 2
    rrw = rbd1.risk_reduction_worth(t)
    assert len(rrw) == 3
    assert (
        pytest.approx(rbd1.ff(t) / rbd1.reliability["valve"].ff(t))
        == rrw["pump1"]
    )
    assert (
        pytest.approx(
            pytest.approx(rbd1.ff(t) / rbd1.reliability["valve"].ff(t))
        )
        == rrw["pump2"]
    )
    assert (
        pytest.approx(
            rbd1.ff(t)
            / (
                rbd1.reliability["pump1"].ff(t)
                * rbd1.reliability["pump2"].ff(t)
            )
        )
        == rrw["valve"]
    )


# Test criticality_importance()
def test_rbd_criticality_importance(rbd1: RBD):
    t = 2
    criticality_importance = rbd1.criticality_importance(t)
    assert len(criticality_importance) == 3
    assert (
        pytest.approx(
            # Birnbaum importance:
            (
                rbd1.reliability["valve"].sf(t)
                - rbd1.reliability["pump2"].sf(t)
                * rbd1.reliability["valve"].sf(t)
            )
            # Correction factor:
            * (rbd1.reliability["pump1"].sf(t) / rbd1.sf(t))
        )
        == criticality_importance["pump1"]
    )
    assert (
        pytest.approx(
            # Birnbaum importance:
            (
                rbd1.reliability["valve"].sf(t)
                - rbd1.reliability["pump1"].sf(t)
                * rbd1.reliability["valve"].sf(t)
            )
            # Correction factor:
            * (rbd1.reliability["pump2"].sf(t) / rbd1.sf(t))
        )
        == criticality_importance["pump2"]
    )
    assert (
        pytest.approx(
            # Birnbaum importance:
            (
                1
                - rbd1.reliability["pump1"].ff(t)
                * rbd1.reliability["pump2"].ff(t)
            )
            # Correction factor:
            * (rbd1.reliability["valve"].sf(t) / rbd1.sf(t))
        )
        == criticality_importance["valve"]
    )


# Test fussel_vesely() w/ cut-set method


def test_fussel_vesely_incorrect_fv_type(rbd1: RBD):
    t = 2
    with pytest.raises(ValueError):
        rbd1.fussel_vesely(t, fv_type="a")


def test_fussel_vesely_c_rbd1(rbd1: RBD):
    t = 2
    fv_importance = rbd1.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(
            rbd1.reliability["pump1"].ff(t)
            * rbd1.reliability["pump2"].ff(t)
            / rbd1.ff(t)
        )
        == fv_importance["pump1"]
    )
    assert (
        pytest.approx(
            rbd1.reliability["pump1"].ff(t)
            * rbd1.reliability["pump2"].ff(t)
            / rbd1.ff(t)
        )
        == fv_importance["pump2"]
    )
    assert (
        pytest.approx(rbd1.reliability["valve"].ff(t) / rbd1.ff(t))
        == fv_importance["valve"]
    )


def test_fussel_vesely_c_series(rbd_series: RBD):
    t = 2
    fv_importance = rbd_series.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(rbd_series.reliability[2].ff(t) / rbd_series.ff(t))
        == fv_importance[2]
    )
    assert (
        pytest.approx(rbd_series.reliability[3].ff(t) / rbd_series.ff(t))
        == fv_importance[3]
    )
    assert (
        pytest.approx(rbd_series.reliability[4].ff(t) / rbd_series.ff(t))
        == fv_importance[4]
    )


def test_fussel_vesely_c_parallel(rbd_parallel: RBD):
    t = 2
    fv_importance = rbd_parallel.fussel_vesely(t, fv_type="c")
    fv_expected = (
        rbd_parallel.reliability[2].ff(t)
        * rbd_parallel.reliability[3].ff(t)
        * rbd_parallel.reliability[4].ff(t)
        / rbd_parallel.ff(t)
    )
    assert pytest.approx(fv_expected) == fv_importance[2]
    assert pytest.approx(fv_expected) == fv_importance[3]
    assert pytest.approx(fv_expected) == fv_importance[4]


def test_fussel_vesely_c_rbd2(rbd2: RBD):
    t = 2
    fv_importance = rbd2.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(rbd2.reliability[2].ff(t) / rbd2.ff(t))
        == fv_importance[2]
    )
    assert (
        pytest.approx(
            rbd2.reliability[3].ff(t) * rbd2.reliability[4].ff(t) / rbd2.ff(t)
        )
        == fv_importance[3]
    )
    assert (
        pytest.approx(
            (
                rbd2.reliability[4].ff(t) * rbd2.reliability[5].ff(t)
                + rbd2.reliability[3].ff(t) * rbd2.reliability[4].ff(t)
                + rbd2.reliability[4].ff(t) * rbd2.reliability[6].ff(t)
            )
            / rbd2.ff(t)
        )
        == fv_importance[4]
    )
    assert (
        pytest.approx(
            rbd2.reliability[4].ff(t) * rbd2.reliability[5].ff(t) / rbd2.ff(t)
        )
        == fv_importance[5]
    )
    assert (
        pytest.approx(
            rbd2.reliability[4].ff(t) * rbd2.reliability[6].ff(t) / rbd2.ff(t)
        )
        == fv_importance[6]
    )
    assert (
        pytest.approx(rbd2.reliability[7].ff(t) / rbd2.ff(t))
        == fv_importance[7]
    )


def test_fussel_vesely_c_rbd3(rbd3: RBD):
    t = 2
    fv_importance = rbd3.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(
            (
                rbd3.reliability[1].ff(t) * rbd3.reliability[3].ff(t)
                + rbd3.reliability[1].ff(t)
                * rbd3.reliability[4].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[1]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[2].ff(t) * rbd3.reliability[4].ff(t)
                + rbd3.reliability[2].ff(t)
                * rbd3.reliability[3].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[2]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[1].ff(t) * rbd3.reliability[3].ff(t)
                + rbd3.reliability[2].ff(t)
                * rbd3.reliability[3].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[3]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[2].ff(t) * rbd3.reliability[4].ff(t)
                + rbd3.reliability[1].ff(t)
                * rbd3.reliability[4].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[4]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[1].ff(t)
                * rbd3.reliability[4].ff(t)
                * rbd3.reliability[5].ff(t)
                + rbd3.reliability[2].ff(t)
                * rbd3.reliability[3].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[5]
    )


def test_fussel_vesely_c_repeated_component_parallel(
    rbd_repeated_component_parallel: RBD,
):
    rbd = rbd_repeated_component_parallel
    t = 2
    fv_importance = rbd.fussel_vesely(t, fv_type="c")
    fv_expected = (
        rbd.reliability[2].ff(t)
        * rbd.reliability[3].ff(t)
        * rbd.reliability[4].ff(t)
        / rbd.ff(t)
    )
    assert pytest.approx(fv_expected) == fv_importance[2]
    assert pytest.approx(fv_expected) == fv_importance[3]
    assert pytest.approx(fv_expected) == fv_importance[4]
    assert pytest.approx(fv_expected) == fv_importance[5]


def test_fussel_vesely_c_rbd_repeated_component_series(
    rbd_repeated_component_series: RBD,
):
    rbd = rbd_repeated_component_series
    t = 2
    fv_importance = rbd.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(rbd.reliability[2].ff(t) / rbd.ff(t)) == fv_importance[2]
    )
    assert (
        pytest.approx(rbd.reliability[3].ff(t) / rbd.ff(t)) == fv_importance[3]
    )
    assert (
        pytest.approx(rbd.reliability[2].ff(t) / rbd.ff(t)) == fv_importance[4]
    )


def test_fussel_vesely_c_rbd_repeated_component_composite(
    rbd_repeated_component_composite: RBD,
):
    rbd = rbd_repeated_component_composite
    t = 2
    fv_importance = rbd.fussel_vesely(t, fv_type="c")
    assert (
        pytest.approx(
            (
                rbd.reliability[2].ff(t)
                + rbd.reliability[2].ff(t) * rbd.reliability[3].ff(t)
            )
            / rbd.ff(t)
        )
        == fv_importance[2]
    )
    assert (
        pytest.approx(rbd.reliability[2].ff(t) / rbd.ff(t)) == fv_importance[3]
    )
    assert (
        pytest.approx(
            rbd.reliability[2].ff(t) * rbd.reliability[3].ff(t) / rbd.ff(t)
        )
        == fv_importance[4]
    )


# Test fussel_vesely() w/ path-set method


def test_fussel_vesely_p_rbd1(rbd1: RBD):
    t = 2
    fv_importance = rbd1.fussel_vesely(t, fv_type="p")
    assert (
        pytest.approx(
            rbd1.reliability["pump1"].ff(t)
            * rbd1.reliability["valve"].ff(t)
            / rbd1.ff(t)
        )
        == fv_importance["pump1"]
    )
    assert (
        pytest.approx(
            rbd1.reliability["pump2"].ff(t)
            * rbd1.reliability["valve"].ff(t)
            / rbd1.ff(t)
        )
        == fv_importance["pump2"]
    )
    assert (
        pytest.approx(
            (
                rbd1.reliability["pump1"].ff(t)
                * rbd1.reliability["valve"].ff(t)
                + rbd1.reliability["pump2"].ff(t)
                * rbd1.reliability["valve"].ff(t)
            )
            / rbd1.ff(t)
        )
        == fv_importance["valve"]
    )


def test_fussel_vesely_p_series(rbd_series: RBD):
    t = 2
    fv_importance = rbd_series.fussel_vesely(t, fv_type="p")
    expected_fv_importance = (
        rbd_series.reliability[2].ff(t)
        * rbd_series.reliability[3].ff(t)
        * rbd_series.reliability[4].ff(t)
        / rbd_series.ff(t)
    )
    assert pytest.approx(expected_fv_importance) == fv_importance[2]
    assert pytest.approx(expected_fv_importance) == fv_importance[3]
    assert pytest.approx(expected_fv_importance) == fv_importance[4]


def test_fussel_vesely_p_parallel(rbd_parallel: RBD):
    t = 2
    fv_importance = rbd_parallel.fussel_vesely(t, fv_type="p")
    assert (
        pytest.approx(rbd_parallel.reliability[2].ff(t) / rbd_parallel.ff(t))
        == fv_importance[2]
    )
    assert (
        pytest.approx(rbd_parallel.reliability[3].ff(t) / rbd_parallel.ff(t))
        == fv_importance[3]
    )
    assert (
        pytest.approx(rbd_parallel.reliability[4].ff(t) / rbd_parallel.ff(t))
        == fv_importance[4]
    )


def test_fussel_vesely_p_rbd2(rbd2: RBD):
    t = 2
    fv_importance = rbd2.fussel_vesely(t, fv_type="p")
    assert (
        pytest.approx(
            (
                (
                    rbd2.reliability[2].ff(t)
                    * rbd2.reliability[3].ff(t)
                    * rbd2.reliability[5].ff(t)
                    * rbd2.reliability[6].ff(t)
                    * rbd2.reliability[7].ff(t)
                )
                + (
                    rbd2.reliability[2].ff(t)
                    * rbd2.reliability[4].ff(t)
                    * rbd2.reliability[7].ff(t)
                )
            )
            / rbd2.ff(t)
        )
        == fv_importance[2]
    )
    assert (
        pytest.approx(
            (
                rbd2.reliability[2].ff(t)
                * rbd2.reliability[3].ff(t)
                * rbd2.reliability[5].ff(t)
                * rbd2.reliability[6].ff(t)
                * rbd2.reliability[7].ff(t)
            )
            / rbd2.ff(t)
        )
        == fv_importance[3]
    )
    assert (
        pytest.approx(
            (
                rbd2.reliability[2].ff(t)
                * rbd2.reliability[4].ff(t)
                * rbd2.reliability[7].ff(t)
            )
            / rbd2.ff(t)
        )
        == fv_importance[4]
    )
    assert (
        pytest.approx(
            (
                rbd2.reliability[2].ff(t)
                * rbd2.reliability[3].ff(t)
                * rbd2.reliability[5].ff(t)
                * rbd2.reliability[6].ff(t)
                * rbd2.reliability[7].ff(t)
            )
            / rbd2.ff(t)
        )
        == fv_importance[5]
    )
    assert (
        pytest.approx(
            (
                rbd2.reliability[2].ff(t)
                * rbd2.reliability[3].ff(t)
                * rbd2.reliability[5].ff(t)
                * rbd2.reliability[6].ff(t)
                * rbd2.reliability[7].ff(t)
            )
            / rbd2.ff(t)
        )
        == fv_importance[6]
    )
    assert (
        pytest.approx(
            (
                (
                    rbd2.reliability[2].ff(t)
                    * rbd2.reliability[3].ff(t)
                    * rbd2.reliability[5].ff(t)
                    * rbd2.reliability[6].ff(t)
                    * rbd2.reliability[7].ff(t)
                )
                + (
                    rbd2.reliability[2].ff(t)
                    * rbd2.reliability[4].ff(t)
                    * rbd2.reliability[7].ff(t)
                )
            )
            / rbd2.ff(t)
        )
    ) == fv_importance[7]


def test_fussel_vesely_p_rbd3(rbd3: RBD):
    t = 2
    fv_importance = rbd3.fussel_vesely(t, fv_type="p")
    assert (
        pytest.approx(
            (
                (rbd3.reliability[1].ff(t) * rbd3.reliability[2].ff(t))
                + (
                    rbd3.reliability[1].ff(t)
                    * rbd3.reliability[5].ff(t)
                    * rbd3.reliability[4].ff(t)
                )
            )
            / rbd3.ff(t)
        )
        == fv_importance[1]
    )
    assert (
        pytest.approx(
            (
                (rbd3.reliability[1].ff(t) * rbd3.reliability[2].ff(t))
                + (
                    rbd3.reliability[3].ff(t)
                    * rbd3.reliability[5].ff(t)
                    * rbd3.reliability[2].ff(t)
                )
            )
            / rbd3.ff(t)
        )
        == fv_importance[2]
    )
    assert (
        pytest.approx(
            (
                (rbd3.reliability[3].ff(t) * rbd3.reliability[4].ff(t))
                + (
                    rbd3.reliability[3].ff(t)
                    * rbd3.reliability[5].ff(t)
                    * rbd3.reliability[2].ff(t)
                )
            )
            / rbd3.ff(t)
        )
        == fv_importance[3]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[3].ff(t) * rbd3.reliability[4].ff(t)
                + rbd3.reliability[1].ff(t)
                * rbd3.reliability[4].ff(t)
                * rbd3.reliability[5].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[4]
    )
    assert (
        pytest.approx(
            (
                rbd3.reliability[1].ff(t)
                * rbd3.reliability[5].ff(t)
                * rbd3.reliability[4].ff(t)
                + rbd3.reliability[3].ff(t)
                * rbd3.reliability[5].ff(t)
                * rbd3.reliability[2].ff(t)
            )
            / rbd3.ff(t)
        )
        == fv_importance[5]
    )


def test_fussel_vesely_p_repeated_component_parallel(
    rbd_repeated_component_parallel: RBD,
):
    rbd = rbd_repeated_component_parallel
    t = 2
    fv_importance = rbd.fussel_vesely(t, fv_type="p")
    assert (
        pytest.approx(rbd.reliability[2].ff(t) / rbd.ff(t)) == fv_importance[2]
    )
    assert (
        pytest.approx(rbd.reliability[3].ff(t) / rbd.ff(t)) == fv_importance[3]
    )
    assert (
        pytest.approx(rbd.reliability[4].ff(t) / rbd.ff(t)) == fv_importance[4]
    )
    assert (
        pytest.approx(rbd.reliability[2].ff(t) / rbd.ff(t)) == fv_importance[5]
    )


def test_fussel_vesely_p_rbd_repeated_component_series(
    rbd_repeated_component_series: RBD,
):
    rbd = rbd_repeated_component_series
    t = 2
    fv_importance = rbd.fussel_vesely(t, fv_type="p")
    expected_fv_importance = (
        rbd.reliability[2].ff(t) * rbd.reliability[3].ff(t) / rbd.ff(t)
    )
    assert pytest.approx(expected_fv_importance) == fv_importance[2]
    assert pytest.approx(expected_fv_importance) == fv_importance[3]
    assert pytest.approx(expected_fv_importance) == fv_importance[4]



"""
Tests many of the RBD class methods.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""

from repyability.rbd.rbd import RBD


def test_rbd_get_min_path_sets_rbd_series(rbd_series: RBD):
    assert {frozenset([1, 2, 3, 4, 5])} == rbd_series.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_parallel(rbd_parallel: RBD):
    assert {
        frozenset([1, 2, 5]),
        frozenset([1, 3, 5]),
        frozenset([1, 4, 5]),
    } == rbd_parallel.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd1(rbd1: RBD):
    assert {
        frozenset(["source", "pump1", "valve", "sink"]),
        frozenset(["source", "pump2", "valve", "sink"]),
    } == rbd1.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd1_exclude_in_out_node(rbd1: RBD):
    assert {
        frozenset(["pump1", "valve"]),
        frozenset(["pump2", "valve"]),
    } == rbd1.get_min_path_sets(include_in_out_nodes=False)


def test_rbd_get_min_path_sets_rbd2(rbd2: RBD):
    assert {
        frozenset([1, 2, 3, 5, 6, 7, 8]),
        frozenset([1, 2, 4, 7, 8]),
    } == rbd2.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd3(rbd3: RBD):
    assert {
        frozenset([0, 1, 2, 6]),
        frozenset([0, 1, 5, 4, 6]),
        frozenset([0, 3, 4, 6]),
        frozenset([0, 3, 5, 2, 6]),
    } == rbd3.get_min_path_sets()


def test_rbd_get_min_path_sets_rbd_double_parallel(
    rbd_double_parallel_args: dict,
):
    rbd = RBD(**rbd_double_parallel_args)
    assert {
        frozenset(["s", 1, 3, 4, "t"]),
        frozenset(["s", 2, 3, 4, "t"]),
        frozenset(["s", 1, 3, 5, "t"]),
        frozenset(["s", 2, 3, 5, "t"]),
    } == rbd.get_min_path_sets()



"""
Tests miscellaneous cases for the RBD class.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
import pytest

from repyability.rbd.rbd import RBD
from repyability.tests.fixed_probability import FixedProbabilityFitter


# Check components are correct lengths
def test_rbd_components(rbd1: RBD, rbd2: RBD):
    assert len(rbd1.reliability) == 5
    assert len(rbd2.reliability) == 8


# Test get_all_path_sets()
def test_rbd_get_all_path_sets(rbd1: RBD, rbd2: RBD):
    assert list(rbd1.get_all_path_sets()) == [
        ["source", "pump1", "valve", "sink"],
        ["source", "pump2", "valve", "sink"],
    ]
    assert list(rbd2.get_all_path_sets()) == [
        [1, 2, 3, 5, 6, 7, 8],
        [1, 2, 4, 7, 8],
    ]


def test_rbd_node_type_dict(rbd1: RBD):
    assert rbd1.G.nodes["source"]["type"] == "input_node"
    assert rbd1.G.nodes["pump1"]["type"] == "node"
    assert rbd1.G.nodes["pump2"]["type"] == "node"
    assert rbd1.G.nodes["sink"]["type"] == "output_node"


# Check ValueError's


def test_rbd_node_not_in_edge_list():
    with pytest.raises(ValueError):
        nodes = {1: "input_node", 2: 2, 3: "output_node"}
        edges = [(1, 3)]
        components = {2: FixedProbabilityFitter.from_params(0.8)}
        RBD(nodes, components, edges)


def test_rbd_component_not_in_reliability_dict():
    with pytest.raises(ValueError):
        nodes = {1: "input_node", 2: 2, 3: "output_node"}
        edges = [(1, 2), (2, 3)]
        components = {}
        RBD(nodes, components, edges)


def test_rbd_node_not_in_node_list():
    with pytest.raises(ValueError):
        nodes = {1: "input_node", 3: "output_node"}
        edges = [(1, 2), (2, 3)]
        components = {2: FixedProbabilityFitter.from_params(0.8)}
        RBD(nodes, components, edges)


def test_rbd_node_with_no_output():
    with pytest.raises(ValueError):
        nodes = {1: "input_node", 2: 2, 4: 2, 3: "output_node"}
        edges = [(1, 2), (2, 3), (2, 4)]
        components = {2: FixedProbabilityFitter.from_params(0.8)}
        RBD(nodes, components, edges)


def test_rbd_node_with_no_input():
    with pytest.raises(ValueError):
        nodes = {1: "input_node", 2: 2, 4: 2, 3: "output_node"}
        edges = [(1, 2), (2, 3), (4, 2)]
        components = {2: FixedProbabilityFitter.from_params(0.8)}
        RBD(nodes, components, edges)


def test_rbd_circular_dependency():
    with pytest.raises(ValueError):
        nodes = {"s": "input_node", 2: 2, 3: 2, 4: 2, "t": "output_node"}
        edges = [("s", 2), (2, 3), (3, 4), (4, 2), (4, "t")]
        components = {2: FixedProbabilityFitter.from_params(0.8)}
        RBD(nodes, components, edges)



from itertools import combinations
from typing import Hashable

import networkx as nx


def min_cut_sets(
    G: nx.DiGraph, input_node: Hashable, output_node: Hashable
) -> set[frozenset[Hashable]]:
    """Returns the set of minimal cut-sets of a graph

    Note: doesn't consider k-out-of-n nodes

    Parameters
    ----------
    G : nx.DiGraph
        The graph in question
    input_node : Hashable
        The input/source node
    output_node : Hashable
        The output/sink node

    Returns
    -------
    set[frozenset[Hashable]]
        The set of minimal nodal cut-sets, i.e. if any of these node-sets
        fail, then the system fails
    """
    # Get all possible node subsets (excluding input and output nodes)
    intermediate_nodes = set(G.nodes()) - {input_node, output_node}
    node_sets = get_all_node_subsets(intermediate_nodes)

    # For each node_subset, create a copy of graph G with these nodes removed
    # If input_node->output_node has no path then add it to the cut_sets set
    # But if there are any supersets of this cut-set, then those supersets
    # are not minimal cut-sets, so remove them

    # The set of cut-sets (not necessarily minimal cut-sets)
    cut_sets: set[frozenset] = set()

    # Check if the node-set is a cut-set
    for node_set in node_sets:
        # Create a copy of G with node_subset nodes removed
        H = G.copy()
        H.remove_nodes_from(node_set)

        # Check if the node-set is a cut-set
        if not nx.has_path(H, input_node, output_node):
            # If it is, check that there are no supersets of it in cut_sets
            # because if there are they are definitely not minimal cut-sets
            # and should be removed
            for cut_set in cut_sets.copy():
                if node_set.issubset(cut_set):
                    cut_sets.remove(cut_set)

            # Finally add the node-set to the set of cut-sets
            cut_sets.add(node_set)

    # By now all the non-minimal cut-sets have been removed, so its safe to say
    # that cut_sets is the set of all minimal_cut_sets
    minimal_cut_sets = cut_sets

    return minimal_cut_sets


def get_all_node_subsets(nodes: set[Hashable]) -> list[frozenset[Hashable]]:
    node_subsets = []
    for i in range(len(nodes), 0, -1):
        for comb in list(combinations(nodes, i)):
            node_subsets.append(frozenset(comb))
    return node_subsets



import networkx as nx


class RBDGraph(nx.DiGraph):
    """A class that wraps networkx.DiGraph for the purposes of RBDs

    RBDGraph differs from DiGraph in only a few ways:
    - RBDGraph utilises DiGraph's ability to have node attributes, this is
      where information like the k in the k-out-of-n sense is stored for each
      node.

    Default node attribute dict:
    {
        "k": 1,
        "type": "node"  # Can be "node"/"input_node"/"output_node"
    }

    """

    # Function that returns the default node attribute dict
    def default_node_attr_dict(self):
        return {"k": 1, "type": "node"}

    # networkx relies on the subclass (RBDGraph) to set the variable
    # `node_attr_dict_factory` to the default node attribute dict function.
    node_attr_dict_factory = default_node_attr_dict



from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from queue import PriorityQueue
from typing import Any, Hashable, Iterable

import numpy as np
from dd import autoref as _bdd
from tqdm import tqdm

from repyability.rbd.rbd import RBD
from repyability.rbd.rbd_args_check import check_rbd_node_args_complete


# Event class for simulation
@dataclass(order=True)
class Event:
    """Dataclass to hold an event's information. Comparisons are performed
    by time. status=False means the event is a component failure."""

    time: float
    component: Hashable = field(compare=False)
    status: bool = field(compare=False)


class RepairableRBD(RBD):
    def __init__(
        self,
        nodes: dict[Any, Any],
        reliability: dict[Any, Any],
        repairability: dict[Any, Any],
        edges: Iterable[tuple[Hashable, Hashable]],
        k: dict[Any, int] = {},
        mc_samples: int = 10_000,
    ):
        check_rbd_node_args_complete(nodes, reliability, edges, repairability)
        super().__init__(nodes, reliability, edges, k, mc_samples)
        self.repairability = copy(repairability)

        # Compile BDD, sets self.bdd and self.bdd_system_ref. See compile_bdd()
        # docstring for more info.
        self.compile_bdd()

    def is_system_working(self, component_status: dict[Any, bool]) -> bool:
        """Returns a boolean as to whether the system is working given the
        status of the components

        Parameters
        ----------
        component_status : dict[Any, bool]
            Dictionary with all components where
            component_status[component] = True only if the component is
            working, and = False if not working.

        Returns
        -------
        bool
            True if the system is working, otherwise False.
        """
        # This function uses a Binary Decision Diagram (BDD) to enable
        # efficient component_status -> is_system_working lookup. Something
        # very much required given the rate at which this function is called
        # in the N simulations in availability().

        # Now evaluate the BDD given the component status'
        system_given_component_status = self.bdd.let(
            component_status, self.bdd_system_ref
        )

        system_state = self.bdd.to_expr(system_given_component_status)

        if system_state == "TRUE":
            return True

        # Else False
        return False

    def availability(
        self, t_simulation: float, N: int = 10_000, verbose: bool = False
    ) -> tuple[np.ndarray, np.ndarray]:
        """Returns the times, and availability for those times, as numpy
        arrays

        Parameters
        ----------
        t_simulation : float
            Units of time to run each simulation for
        N : int, optional
            Number of simulations, by default 10_000
        verbose : bool, optional
            If True, displays progress bar of simulations, by default False

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            times, availabilities
        """

        # aggregate_timeline keeps track of how many systems turn on and off
        # at time t.
        # e.g. aggregate_timeline[t] = +2 means two out of the N systems began
        # working again at time t, while aggregate_timeline[t] = -1 means one
        # out of the N systems stopped working at time t.
        # One might expect that due to the majority of distributions being
        # CDFs that generally aggregate_timeline[t] would be only -1 or +1.
        aggregate_timeline: dict[float, int] = defaultdict(lambda: 0)
        aggregate_timeline[t_simulation] = 0

        # Perform N simulations
        for _ in tqdm(
            range(N), disable=not verbose, desc="Running simulations"
        ):

            # 'Turn on' this simulation's system at time t=0
            aggregate_timeline[0] += 1
            curr_system_state = True

            # Keep record of component status', initially they're all working
            component_status: dict[Any, bool] = {
                component: True for component in self.reliability.keys()
            }

            # PriorityQueue supplies failure/repair events in chronological
            # order
            event_queue: PriorityQueue[Event] = PriorityQueue()

            # Get the first failure event for each component
            for component in self.reliability.keys():
                t_event = self.reliability[component].random(1)[0]

                # Only consider it if it occurs within the simulation window
                if t_event < t_simulation:
                    # Event status is False => event is a component failure
                    event_queue.put(Event(t_event, component, False))

            # It is implemented such that no events that occur after the
            # end-time of the simulation are added to the queue, so we just
            # need to keep going through the queue until it's empty
            while not event_queue.empty():
                # Get the next event and update the component's status
                event = event_queue.get()
                component_status[event.component] = event.status

                # Record new system state, it could still be the same as
                # curr_system_state in which case we don't bother changing
                # aggregate_timeline, but if it is different, we need to +/-1
                # if the system has gone on/off-line
                new_system_state = self.is_system_working(component_status)
                if new_system_state != curr_system_state:
                    if new_system_state:
                        aggregate_timeline[event.time] += 1
                    else:
                        aggregate_timeline[event.time] -= 1

                # Set the curr_system_state
                curr_system_state = new_system_state

                # Now we need to get the component's next event
                # If the component just got repaired then we need it's next
                # failure event, otherwise it just broke and we need it's
                # repair event
                if event.status:
                    # Component just got repaired, need next failure event
                    next_event = Event(
                        # Current time (event.time) + time to next failure
                        event.time
                        + self.reliability[event.component].random(1)[0],
                        event.component,
                        False,  # This is a component failure event
                    )
                else:
                    # Component just failed, need its repair event
                    next_event = Event(
                        # Current time (event.time) + time to repair
                        event.time
                        + self.repairability[event.component].random(1)[0],
                        event.component,
                        True,  # This is a component repair event
                    )

                # But only queue up the event if it occurs before the end
                # of the simulation
                if next_event.time < t_simulation:
                    event_queue.put(next_event)

                # Then move on to the next event... until there's no more
                # events before t_simulation

        # Now we need to return the system availability from t=0..t_simulation
        # Using numpy arrays for efficiency
        timeline_arr: np.ndarray = np.array(list(aggregate_timeline.items()))

        # Sort the array by event time
        timeline_arr = timeline_arr[timeline_arr[:, 0].argsort()]
        time = timeline_arr[:, 0]

        # Take the cumulative sum, this is basically calculating for each
        # t just how many systems are working, and divide by N to get
        # availability the as a percentage
        system_availability = timeline_arr[:, 1].cumsum() / N

        return time, system_availability

    def compile_bdd(self):
        """
        Compiles the BDD, setting self.bdd to the BDD manager, and
        self.bdd_system_ref to the BDD variable reference for the system state.
        """
        # Instantiate the manager
        bdd = _bdd.BDD()

        # Enable Dynamic Reordering (Rudell's Sifting Algorithm)
        bdd.configure(reordering=True)

        # For all components, declare the BDD variable,
        # and get the BDD variable reference
        bdd_vars = {}
        for component in self.get_component_names():
            bdd.declare(component)
            bdd_vars[component] = bdd.var(component)

        # Make path expressions
        path_expressions: list[_bdd.Function] = []
        for min_path_set in self.get_min_path_sets(include_in_out_nodes=False):
            min_path_set_as_list = list(min_path_set)
            path_function = bdd_vars[min_path_set_as_list[0]]
            for component in min_path_set_as_list[1:]:
                path_function &= bdd_vars[component]
            path_expressions.append(path_function)

        system: _bdd.Function = path_expressions[0]
        for path_expression in path_expressions[1:]:
            system |= path_expression

        self.bdd = bdd
        self.bdd_system_ref = system

    # Debugging Functions - these help to understand the BDD structure
    def bdd_to_string(self, filename: str) -> str:
        """Returns the BDD as a string. Simply wraps bdd.to_expr()"""
        return self.bdd.to_expr(self.bdd_system_ref)

    def bdd_to_file(self, filename: str):
        """
        Wraps bdd.dump(). The file type is inferred from the extension
        (case insensitive).

        Note: bdd.dump() depends on the python library `pydot` being installed
        (via pip) and the graphviz' `dot` program being in your PATH.
        Install pydot just with pip: `pip install pydot`,
        and install graphviz, at least on Mac, with `brew install graphviz`.
        Installing with brew will at least make sure `dot` is added to your
        PATH.

        Supported extensions:
        '.p' for Pickle
        '.pdf' for PDF
        '.png' for PNG
        '.svg' for SVG
        '.json' for JSON
        """
        self.bdd.dump(filename, roots=[self.bdd_system_ref])
